#!/usr/bin/env python
# setup.py generated by flit for tools that don't yet use PEP 517

from distutils.core import setup

packages = \
['perky']

package_data = \
{'': ['*']}

setup(name='perky',
      version='0.2.3',
      description='A simple, Pythonic file format.  Same interface as the',
      author='Larry Hastings',
      author_email='larry@hastings.org',
      url='https://github.com/larryhastings/perky/',
      packages=packages,
      package_data=package_data,
      python_requires='>=3.5',
     )
